==> mosdns-4.0.0.tar.gz: mosdns-4.0.0/pkg/bundled_upstream/bundled_upstream.go <==
/*
 * Copyright (C) 2020-2022, IrineSistiana
 *
 * This file is part of mosdns.
 *
 * mosdns is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * mosdns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package bundled_upstream

import (
	"context"
	"errors"
	"github.com/IrineSistiana/mosdns/v4/pkg/query_context"
	"github.com/miekg/dns"
	"go.uber.org/zap"
)

type Upstream interface {
	Exchange(ctx context.Context, q *dns.Msg) (*dns.Msg, error)
	Address() string
	Trusted() bool
}

type BundledUpstream struct {
	us     []Upstream
	logger *zap.Logger
}

// NewBundledUpstream creates a BundledUpstream.
// us must contain at least one Upstream.
// To disable logger, set it to nil.
func NewBundledUpstream(us []Upstream, logger *zap.Logger) *BundledUpstream {
	if len(us) < 1 {
		panic("us must contain at least one Upstream")
	}

	if logger == nil {
		logger = zap.NewNop()
	}
	return &BundledUpstream{us: us, logger: logger}
}

type parallelResult struct {
	r    *dns.Msg
	err  error
	from Upstream
}

func (bu *BundledUpstream) ExchangeParallel(ctx context.Context, qCtx *query_context.Context) (*dns.Msg, error) {
	q := qCtx.Q()
	t := len(bu.us)
	if t == 1 {
		u := bu.us[0]
		r, err := u.Exchange(ctx, q)
		if err != nil {
			return nil, err
		}
		bu.logger.Debug("response received", qCtx.InfoField(), zap.String("from", u.Address()))
		return r, nil
	}

	c := make(chan *parallelResult, t) // use buf chan to avoid blocking.
	qCopy := q.Copy()                  // qCtx is not safe for concurrent use.
	for _, u := range bu.us {
		u := u
		go func() {
			r, err := u.Exchange(ctx, qCopy)
			c <- &parallelResult{
				r:    r,
				err:  err,
				from: u,
			}
		}()
	}

	var candidateErrReply *parallelResult
	for i := 0; i < t; i++ {
		select {
		case res := <-c:
			if res.err != nil {
				bu.logger.Warn("upstream failed", qCtx.InfoField(), zap.String("from", res.from.Address()), zap.Error(res.err))
				continue
			}

			if res.r.Rcode == dns.RcodeSuccess || res.from.Trusted() {
				bu.logger.Debug("response accepted", qCtx.InfoField(), zap.String("from", res.from.Address()))
				return res.r, nil
			}

			if candidateErrReply == nil {
				candidateErrReply = res
			}
			bu.logger.Debug("untrusted upstream returned an err rcode", qCtx.InfoField(), zap.String("from", res.from.Address()), zap.Int("rcode", res.r.Rcode))
		case <-ctx.Done():
			return nil, ctx.Err()
		}
	}

	// all upstreams failed or returned an error rcode.
	if candidateErrReply != nil {
		bu.logger.Debug("candidate error response accepted", qCtx.InfoField(), zap.String("from", candidateErrReply.from.Address()))
		return candidateErrReply.r, nil
	}

	return nil, errors.New("no response")
}

==> mosdns-4.5.1.tar.gz: mosdns-4.5.1/pkg/bundled_upstream/bundled_upstream.go <==
/*
 * Copyright (C) 2020-2022, IrineSistiana
 *
 * This file is part of mosdns.
 *
 * mosdns is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * mosdns is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package bundled_upstream

import (
	"context"
	"errors"
	"github.com/IrineSistiana/mosdns/v4/pkg/query_context"
	"github.com/miekg/dns"
	"go.uber.org/zap"
)

type Upstream interface {
	// Exchange sends q to the upstream and waits for response.
	// If any error occurs. Implements must return a nil msg with a non nil error.
	// Otherwise, Implements must a msg with nil error.
	Exchange(ctx context.Context, q *dns.Msg) (*dns.Msg, error)

	// Trusted indicates whether this Upstream is trusted/reliable.
	// If true, responses from this Upstream will be accepted without checking its rcode.
	Trusted() bool

	Address() string
}

type parallelResult struct {
	r    *dns.Msg
	err  error
	from Upstream
}

var nopLogger = zap.NewNop()

var ErrAllFailed = errors.New("all upstreams failed")

func ExchangeParallel(ctx context.Context, qCtx *query_context.Context, upstreams []Upstream, logger *zap.Logger) (*dns.Msg, error) {
	if logger == nil {
		logger = nopLogger
	}

	q := qCtx.Q()
	t := len(upstreams)
	if t == 1 {
		return upstreams[0].Exchange(ctx, q)
	}

	c := make(chan *parallelResult, t) // use buf chan to avoid blocking.
	qCopy := q.Copy()                  // qCtx is not safe for concurrent use.
	for _, u := range upstreams {
		u := u
		go func() {
			r, err := u.Exchange(ctx, qCopy)
			c <- &parallelResult{
				r:    r,
				err:  err,
				from: u,
			}
		}()
	}

	for i := 0; i < t; i++ {
		select {
		case res := <-c:
			if res.err != nil {
				logger.Warn("upstream err", qCtx.InfoField(), zap.String("addr", res.from.Address()))
				continue
			}

			if res.r == nil {
				continue
			}

			if res.r.Rcode != dns.RcodeSuccess {
				if res.from.Trusted() {
					return res.r, nil
				}
				continue
			}

		case <-ctx.Done():
			return nil, ctx.Err()
		}
	}
	return nil, ErrAllFailed
}
